#include <pcl/io/pcd_io.h>
#include <pcl/io/ply_io.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/console/parse.h>
#include <pcl/registration/icp.h>
#include <pcl/registration/ia_ransac.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/common/common_headers.h>
#include <pcl/features/normal_3d.h>
#include <pcl/filters/passthrough.h>
#include <pcl/filters/extract_indices.h>

#include <pcl/registration/correspondence_rejection.h>
#include <pcl/registration/correspondence_estimation.h>

#include <boost/thread/thread.hpp>

#include <iostream>
#include <cstring>
#include <string>

#include <pcl/search/kdtree.h>

#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/ModelCoefficients.h>
 
#include <Eigen/Geometry>
#include <Eigen/Dense>
 
// Schrittweite für manuelle Transformationen und Faktor um diese zu Skalieren 
double schrittweite = 1.0;
double factor = 0.5;

// Hintergrundfarbe in rgb von 0 bis 1 
double background_r = 1.0;
double background_g = 1.0;
double background_b = 1.0;

// Grenzen für PassThrough Filter
double filterLimitMinX = 0.0;
double filterLimitMaxX = 1.0;
double filterLimitMinY = 0.0;
double filterLimitMaxY = 1.0;
double filterLimitMinZ = 0.0;
double filterLimitMaxZ = 1.0;

// benötigt um Speicherort für Ergebnisdatei  anzugeben
std::string saveDestination;

// benötigt zum einmaligen Wechsel nach Ausführung von addPointCloud zu updatePointCloud
bool cloudAddedA = false;
bool cloudAddedB = false;
bool colorize = true;
bool undoEnabled = true;
// Transformationsmatrizen
Eigen::Matrix4f global_transformation_matrix = Eigen::Matrix4f::Identity(); 
Eigen::Matrix4f local_transformation_matrix = Eigen::Matrix4f::Identity();
Eigen::Matrix4f backup_transformation_matrix = Eigen::Matrix4f::Identity();

// Pointclouds für scene- und model-Scans + Hilfspointcloud für Transformationen
pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZRGB> ());
pcl::PointCloud<pcl::PointXYZRGB>::Ptr unmodified_original_cloud (new pcl::PointCloud<pcl::PointXYZRGB> ());
pcl::PointCloud<pcl::PointXYZRGB>::Ptr backup_cloud (new pcl::PointCloud<pcl::PointXYZRGB> ());
pcl::PointCloud<pcl::PointXYZRGB>::Ptr transformed_cloud (new pcl::PointCloud<pcl::PointXYZRGB> ());

// Pointclouds um ausgewählte Punkte zu speichern 
pcl::PointCloud<pcl::PointXYZRGB>::Ptr clicked_points (new pcl::PointCloud<pcl::PointXYZRGB> ());
 
pcl::StopWatch timer;

int mode = 0;

double winkel = 180.0;

int point_size = 1;
int cube_line_width = 2;

bool legendOFF = true;

// Diese Funktion gibt die Hilfe aus 
void showHelp(char * program_name)
{
	std::cout << "\n############################################################################" <<std::endl;
	std::cout << "################ Transformieren und Zuschneiden von Punktwolken ############"<< std::endl;
	std::cout << "############################################################################" <<std::endl;
	std::cout << "\nUsage: " << program_name << " cloud_filename.[pcd]" << " {-nc}"<< std::endl;
	std::cout << "-nc for not colorizing height of scene cloud "<< std::endl;
	std::cout << "\n##########################################################################" <<std::endl;
	std::cout << "Key:              |Symbol:   |Funktion:\n"
            		<< "==================|=========================================================\n"
					<< "Modus: manuelle Transformation: \n"
            		<< "----------------------------------------------------------------------------\n"
            		<< "Numpad 6          |bewegt die Punktwolke um eine Schrittweite in\n                  | die positive Richtung der x-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Numpad 4          |bewegt die Punktwolke um eine Schrittweite in\n                  | die negative Richtung der x-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Numpad 9          |bewegt die Punktwolke um eine Schrittweite in\n                  | die positive Richtung der y-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Numpad 3          |bewegt die Punktwolke um eine Schrittweite in\n                  | die negative Richtung der y-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Numpad 8          |bewegt die Punktwolke um eine Schrittweite in\n                  | die positive Richtung der z-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Numpad 2          |bewegt die Punktwolke um eine Schrittweite in\n                  | die negative Richtung der z-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Shift + Numpad 6  |rotiert die Punktwolke um eine Schrittweite in\n                  | die positive Drehrichtung um die x-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
	    			<< "Shift + Numpad 4  |rotiert die Punktwolke um eine Schrittweite in\n                  | die negative Drehrichtung um die x-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
	    			<< "Shift + Numpad 9  |rotiert die Punktwolke um eine Schrittweite in\n                  | die positive Drehrichtung um die y-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
	    			<< "Shift + Numpad 3  |rotiert die Punktwolke um eine Schrittweite in\n                  | die negative Drehrichtung um die y-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
	    			<< "Shift + Numpad 8  |rotiert die Punktwolke um eine Schrittweite in\n                  | die positive Drehrichtung um die z-Achse\n"
            		<< "------------------|---------------------------------------------------------\n"
	    			<< "Shift + Numpad 2  |rotiert die Punktwolke um eine Schrittweite in\n                  | die negative Drehrichtung um die z-Achse\n"
	    			<< "==================|=========================================================\n"
					<< "Modus: Ausschneiden mit Kubus: \n"
					<< "----------------------------------------------------------------------------\n"
					<< "Numpad 6          |vergrößert den Kubus um eine Schrittweite in\n                  | die positive Richtung der x-Achse\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Numpad 4          |vergrößert den Kubus um eine Schrittweite in\n                  | die negative Richtung der x-Achse\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Numpad 9          |vergrößert den Kubus um eine Schrittweite in\n                  | die positive Richtung der y-Achse\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Numpad 3          |vergrößert den Kubus um eine Schrittweite in\n                  | die negative Richtung der y-Achse\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Numpad 8          |vergrößert den Kubus um eine Schrittweite in\n                  | die positive Richtung der z-Achse\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Numpad 2          |vergrößert den Kubus um eine Schrittweite in\n                  | die negative Richtung der z-Achse\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Shift + Numpad    |verkleinert den Kubus um eine Schrittweite \n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Numpad 5          |Schrittweite verkleinern (/2)\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "Shift + Numpad 5  |Schrittweite vergrößern (*2)\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "b                 |Punkte außerhalb des Kubus entfernen\n"
					<< "------------------|---------------------------------------------------------\n"
					<< "B                 |Punkte innerhalb des Kubus entfernen\n"
            		<< "==================|=========================================================\n"
					<< "Unterstützung durch Punkte: \n"
            		<< "------------------|---------------------------------------------------------\n"
	    			<< "Shift + leftclick |Punkt auswählen\n"
            		<< "------------------|---------------------------------------------------------\n"
	    			<< "n                 |Punktwolke ausrichten (vorher einen Punkt auf Boden\n                  | auswählen)\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Delete            |markierte Punkte bei Bedarf löschen\n"
            		<< "==================|=========================================================\n"
					<< "Allgemein: \n"
            		<< "----------------------------------------------------------------------------\n"            		
            		<< "h                 |diese Hilfe öffnen\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "l                 |Legende ein- und ausblenden\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "m                 |Modus zwischen schneiden und transformieren wechseln\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "t                 |bisherige Transformation ausgeben\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "T                 |bisherige Transformation in einer Zeile \n                  | ausgeben\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "X                 |Transformation in einer Zeile eingeben \n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "x                 |eingegebene Transformation ausführen \n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Backspace         |Filter einmal rückgängig machen\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "Numpad Enter      |Punktwolke speichern\n"
            		<< "------------------|---------------------------------------------------------\n"
            		<< "\n\n" << std::endl;
}


void colorHeight(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud){
	timer.reset();
	// minimum
	float min= 9999.9;
	// maximum
	float max= -9999.9;
	for(size_t i = 0; i < cloud->points.size(); i++){
		if (cloud->points[i].z < min){
			min = cloud->points[i].z;
		}
		if (cloud->points[i].z > max){
			max = cloud->points[i].z;
		}
	}
	float dist = max - min;
	//std::cout << min << " " << max << std::endl;
	for(size_t i = 0; i < cloud->points.size(); i++){
//		cloud->points[i].rgb = (cloud->points[i].z - min ) / dist ;
		// +0.1 * 230 um nicht zu viel im schwarzen bereich zu haben
		cloud->points[i].g = static_cast<uint8_t>((((cloud->points[i].z - min )/dist )+ 0.1 )* 230.0) ;
	}
	//std::cout << "zeit: " << timer.getTime() << std::endl;
}

// printed bei Tastendruck das Symbol der gedrückten Taste auf der Console aus
void printKeySym(const pcl::visualization::KeyboardEvent &event)
{
	std::string str;
	str = event.getKeySym ();
	printf("Taste = %s \n",str.c_str());
}

// Point Picking Callback Funktion zum Auswählen von cloud und model_cloud Punkten, um beide clouds in einem anderen Schritt per ICP angleichen zu können
void pp_callback (const pcl::visualization::PointPickingEvent& event, void* viewarg)
{
	// viewer void pointer casten
	boost::shared_ptr<pcl::visualization::PCLVisualizer> view = *static_cast<boost::shared_ptr<pcl::visualization::PCLVisualizer> *> (viewarg);
	// return bei Fehler 
	if (event.getPointIndex () == -1)
		return;
	// ausgewählten Punkt in current_point zwischenspeichern
	pcl::PointXYZRGB current_point;
	event.getPoint(current_point.x, current_point.y, current_point.z);

	// ausgewählten Punkt den scene_points zuordnen
	clicked_points->points.push_back(current_point);
	// Zeige den markierten cloud Punkt vergrößert in rot an
	pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZRGB> red (clicked_points, 255, 0, 0);
	view->removePointCloud("clicked_points");
	view->addPointCloud(clicked_points, red, "clicked_points");
	view->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 10, "clicked_points");
	
	filterLimitMinX = 9999.9;
	filterLimitMaxX = -9999.9;
	filterLimitMinY = 9999.9;
	filterLimitMaxY = -9999.9;
	filterLimitMinZ = 9999.9;
	filterLimitMaxZ = -9999.9;

	for(size_t i = 0; i < clicked_points->points.size(); i++){
		if (clicked_points->points[i].x < filterLimitMinX){
			filterLimitMinX = clicked_points->points[i].x;
		}
		if (clicked_points->points[i].x > filterLimitMaxX){
			filterLimitMaxX = clicked_points->points[i].x;
		}

		if (clicked_points->points[i].y < filterLimitMinY){
			filterLimitMinY = clicked_points->points[i].y;
		}
		if (clicked_points->points[i].y > filterLimitMaxY){
			filterLimitMaxY = clicked_points->points[i].y;
		}

		if (clicked_points->points[i].z < filterLimitMinZ){
			filterLimitMinZ = clicked_points->points[i].z;
		}
		if (clicked_points->points[i].z > filterLimitMaxZ){
			filterLimitMaxZ = clicked_points->points[i].z;
		}
	}

	// Grenzen des Filters visualisieren
	view->removeShape("passthrough_cube");
	view->addCube(filterLimitMinX, filterLimitMaxX, filterLimitMinY, filterLimitMaxY, filterLimitMinZ, filterLimitMaxZ, 1.0, 0.1, 0.1, "passthrough_cube");
	view->setShapeRenderingProperties(pcl::visualization::PCL_VISUALIZER_LINE_WIDTH, cube_line_width, "passthrough_cube");
}

// Funktion, die bei einer Keyboard Eingabe aufgerufen wird
void keyboardEventOccurred (const pcl::visualization::KeyboardEvent &event, void* viewer_void)
{
	if (undoEnabled && !(event.getKeySym () == "BackSpace") && (event.getKeySym () == "B" || event.getKeySym () == "b" ) && event.keyDown ())
	{
		timer.reset();
		// backup falls man mit dem ausschneiden unzufrieden ist
		pcl::copyPointCloud(*cloud, *backup_cloud);
		backup_transformation_matrix = global_transformation_matrix;
	}
	// viewer void pointer casten
  	boost::shared_ptr<pcl::visualization::PCLVisualizer> viewer = *static_cast<boost::shared_ptr<pcl::visualization::PCLVisualizer> *> (viewer_void);
	
//printKeySym(event);

// Wenn Transformation
	if ( (event.getKeySym () == "KP_Up"     	|| event.getKeySym () == "KP_Down" 	||
			event.getKeySym () == "KP_Right"	|| event.getKeySym () == "KP_Left" 	||
			event.getKeySym () == "KP_Prior" 	|| event.getKeySym () == "KP_Next" 	|| 
			event.getKeySym () == "KP_6"       	|| event.getKeySym () == "KP_4" 	||
			event.getKeySym () == "KP_9" 		|| event.getKeySym () == "KP_3" 	||
			event.getKeySym () == "KP_8" 		|| event.getKeySym () == "KP_2") && event.keyDown ())
	{
		if(mode == 0){
/////////////// Translation //////////////////		
		
			// Transformationsmatrix wieder zurücksetzen, um keine alten Transformationen mit einzubeziehen
			local_transformation_matrix = Eigen::Matrix4f::Identity();
		
			//Numpad 8 : bewegt die Punktwolke um eine Schrittweite in die positive Richtung der z-Achse
			if (event.getKeySym () == "KP_Up" && event.keyDown ())
			{
					local_transformation_matrix (2,3) = schrittweite ;
			}

			//Numpad 2 : bewegt die Punktwolke um eine Schrittweite in die negative Richtung der z-Achse
			if (event.getKeySym () == "KP_Down" && event.keyDown ())
			{
				local_transformation_matrix (2,3) = - schrittweite ;
			}
			
			//Numpad 6 : bewegt die Punktwolke um eine Schrittweite in die positive Richtung der x-Achse
			if (event.getKeySym () == "KP_Right" && event.keyDown ())
			{
				local_transformation_matrix (0,3) = schrittweite ;
			}
		
			//Numpad 4 : bewegt die Punktwolke um eine Schrittweite in die negative Richtung der x-Achse
			if (event.getKeySym () == "KP_Left" && event.keyDown ())
			{
				local_transformation_matrix (0,3) = - schrittweite ;
			}
		
			//Numpad 9 : bewegt die Punktwolke um eine Schrittweite in die positive Richtung der y-Achse
			if (event.getKeySym () == "KP_Prior" && event.keyDown ())
			{
				local_transformation_matrix (1,3) = schrittweite ;
			}

			//Numpad 3 : bewegt die Punktwolke um eine Schrittweite in die negative Richtung der y-Achse
			if (event.getKeySym () == "KP_Next" && event.keyDown ())
			{
				local_transformation_matrix (1,3) = - schrittweite ;
			}
	/////////////// Rotation //////////////////

			//Shift + Numpad 6 : rotiert die Punktwolke um eine Schrittweite in die positive Drehrichtung um die x-Achse
			if (event.getKeySym () == "KP_6" && event.keyDown ())
				{
				// Rotationsmatrix aufstellen
				local_transformation_matrix (1,1) = cos(schrittweite);
				local_transformation_matrix (1,2) = -sin(schrittweite);
				local_transformation_matrix (2,1) = sin(schrittweite);
				local_transformation_matrix (2,2) = cos(schrittweite);
			}

			//Shift + Numpad 4 : rotiert die Punktwolke um eine Schrittweite in die negative Drehrichtung um die x-Achse
			if (event.getKeySym () == "KP_4" && event.keyDown ())
				{
				// Rotationsmatrix aufstellen
				local_transformation_matrix (1,1) = cos(-schrittweite);
				local_transformation_matrix (1,2) = -sin(-schrittweite);
				local_transformation_matrix (2,1) = sin(-schrittweite);
				local_transformation_matrix (2,2) = cos(-schrittweite);
			}

			//Shift + Numpad 9 : rotiert die Punktwolke um eine Schrittweite in die positive Drehrichtung um die y-Achse
			if (event.getKeySym () == "KP_9" && event.keyDown ())
				{
				// Rotationsmatrix aufstellen
				local_transformation_matrix (0,0) = cos(schrittweite);
				local_transformation_matrix (0,2) = sin(schrittweite);
				local_transformation_matrix (2,0) = -sin(schrittweite);
				local_transformation_matrix (2,2) = cos(schrittweite);
			}

			//Shift + Numpad 3 : rotiert die Punktwolke um eine Schrittweite in die negative Drehrichtung um die y-Achse
			if (event.getKeySym () == "KP_3" && event.keyDown ())
				{
				// Rotationsmatrix aufstellen
				local_transformation_matrix (0,0) = cos(-schrittweite);
				local_transformation_matrix (0,2) = sin(-schrittweite);
				local_transformation_matrix (2,0) = -sin(-schrittweite);
				local_transformation_matrix (2,2) = cos(-schrittweite);
			}

			//Shift + Numpad 8 : rotiert die Punktwolke um eine Schrittweite in die positive Drehrichtung um die z-Achse
			if (event.getKeySym () == "KP_8" && event.keyDown ())
				{
				// Rotationsmatrix aufstellen
				local_transformation_matrix (0,0) = cos(schrittweite);
				local_transformation_matrix (0,1) = -sin(schrittweite);
				local_transformation_matrix (1,0) = sin(schrittweite);
				local_transformation_matrix (1,1) = cos(schrittweite);
			}

			//Shift + Numpad 2 : rotiert die Punktwolke um eine Schrittweite in die negative Drehrichtung um die z-Achse
			if (event.getKeySym () == "KP_2" && event.keyDown ())
				{
				// Rotationsmatrix aufstellen
				local_transformation_matrix (0,0) = cos(-schrittweite);
				local_transformation_matrix (0,1) = -sin(-schrittweite);
				local_transformation_matrix (1,0) = sin(-schrittweite);
				local_transformation_matrix (1,1) = cos(-schrittweite);
			}


	///////////////// Transformation anwenden /////////////////

			// gesamte bisherige Transformation in global_transformation_matrix speichern
			global_transformation_matrix = global_transformation_matrix  * local_transformation_matrix;

			// globale Transformation auf model_cloud anwenden und transformierte Cloud in transformed_model_cloud speichern
			pcl::transformPointCloud (*unmodified_original_cloud, *cloud, global_transformation_matrix);
		}else{
			if(mode == 1){
				// Box vergrößern mit 6 und 4 in x Richtung , 9 und 3 für y und 8 und 2 für z
				//Numpad 6
				if (event.getKeySym () == "KP_Right" && event.keyDown ())
				{
					filterLimitMaxX += schrittweite ;
				}

				//Numpad 4
				if (event.getKeySym () == "KP_Left" && event.keyDown ())
				{
					filterLimitMinX -= schrittweite ;
				}

				//Numpad 9
				if (event.getKeySym () == "KP_Prior" && event.keyDown ())
				{
					filterLimitMaxY += schrittweite ;
				}

				//Numpad 3
				if (event.getKeySym () == "KP_Next" && event.keyDown ())
				{
					filterLimitMinY -= schrittweite ;
				}

				//Numpad 8
				if (event.getKeySym () == "KP_Up" && event.keyDown ())
				{
					filterLimitMaxZ += schrittweite ;
				}

				//Numpad 2
				if (event.getKeySym () == "KP_Down" && event.keyDown ())
				{
					filterLimitMinZ -= schrittweite ;
				}

			// Box verkleinern mit Shift 6 und 4 in x Richtung ,shift 9 und 3 für y und shift 8 und 2 für z

				//Shift + Numpad 6
				if (event.getKeySym () == "KP_6" && event.keyDown ())
				{
					filterLimitMaxX -= schrittweite ;
				}

				//Shift + Numpad 4
				if (event.getKeySym () == "KP_4" && event.keyDown ())
				{
					filterLimitMinX += schrittweite ;
				}

				//Shift + Numpad 9
				if (event.getKeySym () == "KP_9" && event.keyDown ())
				{
					filterLimitMaxY -= schrittweite ;
				}

				//Shift + Numpad 3
				if (event.getKeySym () == "KP_3" && event.keyDown ())
				{
					filterLimitMinY += schrittweite ;
				}

				//Shift + Numpad 8
				if (event.getKeySym () == "KP_8" && event.keyDown ())
				{
					filterLimitMaxZ -= schrittweite ;
				}

				//Shift + Numpad 2
				if (event.getKeySym () == "KP_2" && event.keyDown ())
				{
					filterLimitMinZ += schrittweite ;
				}

				// Grenzen des Filters visualisieren
				viewer->removeShape("passthrough_cube");
				viewer->addCube(filterLimitMinX, filterLimitMaxX, filterLimitMinY, filterLimitMaxY, filterLimitMinZ, filterLimitMaxZ, 1.0, 0.1, 0.1, "passthrough_cube");
				viewer->setShapeRenderingProperties(pcl::visualization::PCL_VISUALIZER_LINE_WIDTH, cube_line_width, "passthrough_cube");

				// Grenzen ausgeben
				//cout << "filterLimitMinX = " << filterLimitMinX << endl << "filterLimitMaxX = " << filterLimitMaxX << endl << "filterLimitMinY = "
				//		<< filterLimitMinY << endl << "filterLimitMaxY = " << filterLimitMaxY << endl << "filterLimitMinZ = " << filterLimitMinZ <<
				//		endl << "filterLimitMaxZ = " << filterLimitMaxZ << endl << endl;
			}
		}
	}


	if (event.getKeySym () == "x" && event.keyDown ())
	{
		// globale Transformation auf model_cloud anwenden und transformierte Cloud in transformed_model_cloud speichern
		pcl::transformPointCloud (*unmodified_original_cloud, *cloud, global_transformation_matrix);
	}
///////////////// sonstige Funktionen /////////////////

	//Event bei Numpad 5 - Schrittweite verkleinern um Faktor 0.1
	if (event.getKeySym () == "KP_Begin" && event.keyDown ())
  	{
		schrittweite = schrittweite * factor;

		viewer->removeShape("stepsize");
		std::stringstream stepsizeStream;
		stepsizeStream << "stepsize: " << schrittweite;
		viewer->addText(stepsizeStream.str(), 10, 30, 16, 0.8, 0.0, 0.0, "stepsize");
  	}
  	//Event bei Shift + Numpad 5 - Schrittweite vergrößern um Faktor 10
  	if (event.getKeySym () == "KP_5" && event.keyDown ())
  	{
		schrittweite = schrittweite / factor;

		viewer->removeShape("stepsize");
		std::stringstream stepsizeStream;
		stepsizeStream << "stepsize: " << schrittweite;
		viewer->addText(stepsizeStream.str(), 10, 30, 16, 0.8, 0.0, 0.0, "stepsize");
  	}
  	
  	//Event bei h - Hilfe öffnen in Konsole
	if (event.getKeySym () == "h" && event.keyDown ())
  	{
		showHelp ("./pointcloud_editor");
  	}
  	
  	//Event bei t - Transformationsmatrix in Konsole anzeigen
	if (event.getKeySym () == "t" && event.keyDown ())
  	{  	
	  	// Ausgabe der Transformation
		printf ("\nTransformationsmatrix :\n");
		printf ("    | %+f %+f %+f %+f | \n", global_transformation_matrix (0,0), global_transformation_matrix (0,1), global_transformation_matrix (0,2), global_transformation_matrix (0,3));
		printf ("    | %+f %+f %+f %+f | \n", global_transformation_matrix (1,0), global_transformation_matrix (1,1), global_transformation_matrix (1,2), global_transformation_matrix (1,3));
		printf ("    | %+f %+f %+f %+f | \n", global_transformation_matrix (2,0), global_transformation_matrix (2,1), global_transformation_matrix (2,2), global_transformation_matrix (2,3));
		printf ("    |  0         0         0         1        | \n");
	}
	
  	//Event bei t - Transformationsmatrix in Konsole anzeigen
	if (event.getKeySym () == "T" && event.keyDown ())
  	{
	  	// Ausgabe der Transformation
		printf ("\nTransformationsmatrix :\n");
		std::cout << global_transformation_matrix (0,0) <<
				" " << global_transformation_matrix (0,1) <<
				" " << global_transformation_matrix (0,2) <<
				" " << global_transformation_matrix (0,3) <<
				" " << global_transformation_matrix (1,0) <<
				" " << global_transformation_matrix (1,1) <<
				" " << global_transformation_matrix (1,2) <<
				" " << global_transformation_matrix (1,3) <<
				" " << global_transformation_matrix (2,0) <<
				" " << global_transformation_matrix (2,1) <<
				" " << global_transformation_matrix (2,2) <<
				" " << global_transformation_matrix (2,3) <<
				" " << global_transformation_matrix (3,0) <<
				" " << global_transformation_matrix (3,1) <<
				" " << global_transformation_matrix (3,2) <<
				" " << global_transformation_matrix (3,3) << std::endl;
	}

  	//Event bei t - Transformationsmatrix in Konsole anzeigen
	if (event.getKeySym () == "X" && event.keyDown ())
  	{
		cin >> global_transformation_matrix (0,0) >>
				global_transformation_matrix (0,1) >>
				 global_transformation_matrix (0,2) >>
				 global_transformation_matrix (0,3) >>
				 global_transformation_matrix (1,0) >>
				 global_transformation_matrix (1,1) >>
				 global_transformation_matrix (1,2) >>
				 global_transformation_matrix (1,3) >>
				 global_transformation_matrix (2,0) >>
				 global_transformation_matrix (2,1) >>
				 global_transformation_matrix (2,2) >>
				 global_transformation_matrix (2,3) >>
				 global_transformation_matrix (3,0) >>
				 global_transformation_matrix (3,1) >>
				 global_transformation_matrix (3,2) >>
				 global_transformation_matrix (3,3) ;

		std::cout << "input was \n" << global_transformation_matrix << std::endl;
		std::cout << "press x to use this transform" << std::endl;
	}

  	//Event bei delete bzw Entf - alle ausgewählten Punkte löschen
	if (event.getKeySym () == "Delete" && event.keyDown ())
  	{  	
		// alle ausgewählten Punkte aus Viewer entfernen
  		viewer->removePointCloud("clicked_points");
  		// alle ausgewählten Punkte löschen
		clicked_points->points.clear();
  	}

  	
	//Event bei Backspace - letztes ausschneiden rückgängig machen
  	if (event.getKeySym () == "BackSpace" && event.keyDown ())
  	{
		pcl::copyPointCloud (*backup_cloud, *cloud);
		global_transformation_matrix = backup_transformation_matrix;
  	}

	//Event bei b - PassThrough Filter anwenden
	if (event.getKeySym() == "b" && event.keyDown()) {
		pcl::PassThrough<pcl::PointXYZRGB> pass;
		// Filter auf X-Achse anwenden
		pass.setInputCloud(cloud);
		pass.setFilterFieldName("x");
		pass.setFilterLimits(filterLimitMinX, filterLimitMaxX);
		pass.filter(*cloud);

		// Filter auf Y-Achse anwenden
		pass.setInputCloud(cloud);
		pass.setFilterFieldName("y");
		pass.setFilterLimits(filterLimitMinY, filterLimitMaxY);
		pass.filter(*cloud);

		//Filter auf Z-Achse anwenden
		pass.setInputCloud(cloud);
		pass.setFilterFieldName("z");
		pass.setFilterLimits(filterLimitMinZ, filterLimitMaxZ);
		pass.filter(*cloud);

		Eigen::Matrix4f tmp_inv_global_transform = global_transformation_matrix.inverse();
		pcl::transformPointCloud (*cloud, *unmodified_original_cloud, tmp_inv_global_transform);

		// Grenzwürfel entfernen
		viewer->removeShape("passthrough_cube");
	}

	//Event bei B - inversen PassThrough Filter anwenden
	if (event.getKeySym() == "B" && event.keyDown()) {
		pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_temp_filter(new pcl::PointCloud<pcl::PointXYZRGB>);

		for(size_t filter_i = 0; filter_i < cloud->points.size(); filter_i++){
			if(!((cloud->points[filter_i].x > filterLimitMinX) && (cloud->points[filter_i].x < filterLimitMaxX) && (cloud->points[filter_i].y > filterLimitMinY) && (cloud->points[filter_i].y < filterLimitMaxY) && (cloud->points[filter_i].z > filterLimitMinZ) && (cloud->points[filter_i].z < filterLimitMaxZ))){
				cloud_temp_filter->points.push_back(cloud->points[filter_i]);
			}
		}
		cloud_temp_filter->height = 1;
		cloud_temp_filter->width = cloud_temp_filter->points.size();

		*cloud = *cloud_temp_filter;

		Eigen::Matrix4f tmp_inv_global_transform = global_transformation_matrix.inverse();
		pcl::transformPointCloud (*cloud, *unmodified_original_cloud, tmp_inv_global_transform);

		// Grenzwürfel entfernen
		viewer->removeShape("passthrough_cube");
	}

	//Event bei n - Cloud mit Boden in x-y-Ebene drehen und Ursprung auf Boden setzen
	if (event.getKeySym() == "n" && event.keyDown() && clicked_points->points.size() > 0) {
		int counter = 0;
		while (((winkel * 180 / 3.14) > 1.0) && (counter < 20)){
			// Ebene bestimmen
			// erzeuge kd tree für inputs
			pcl::KdTreeFLANN< pcl::PointXYZRGB> kdtree;
			kdtree.setInputCloud (cloud);
			// Vektor um die Indices der gefundenen Nachbarn im Umkreis radius des searchPoint zu speichern
			std::vector<int> pointIdxRadiusSearch;
			// Vektor um die Distanzen der gefundenen Nachbarn im Umkreis radius des searchPoint zu speichern
			std::vector<float> pointRadiusSquaredDistance;

			// Suche nach nächsten Nachbarn im Umkreis radius, Rückgabewert = Anzahl gefundener Nachbarn
			kdtree.radiusSearch (clicked_points->points[0], 0.3, pointIdxRadiusSearch, pointRadiusSquaredDistance);
			pcl::PointCloud< pcl::PointXYZRGB>::Ptr local_cloud (new pcl::PointCloud< pcl::PointXYZRGB>);

			for (size_t iz = 0; iz < pointIdxRadiusSearch.size (); ++iz) {
				// gefundene Nachbarn in local_cloud speichern
				local_cloud->points.push_back(cloud->points[ pointIdxRadiusSearch[iz] ] );
			}
			// Ebene finden
			pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);
			pcl::PointIndices::Ptr inliers (new pcl::PointIndices);
			// Erstellung des Objekts zur Segmentation, einer Ebene
			pcl::SACSegmentation<pcl::PointXYZRGB> seg;
			// Optional
			seg.setOptimizeCoefficients (true);
			// Obligatorisch

			// zu suchendes Modell auf Ebene setzen
			seg.setModelType (pcl::SACMODEL_PLANE);
			seg.setMethodType (pcl::SAC_RANSAC);
			// distance threshold für Ebene festlegen
			seg.setDistanceThreshold (0.005);
			// Input
			seg.setInputCloud (local_cloud);
			// segmentieren
			seg.segment (*inliers, *coefficients);
			// mit gefundener Ebene arbeiten
			if (inliers->indices.size () == 0)
			{
				PCL_ERROR ("Could not estimate a planar model for the given dataset.");
			}else{
				// Winkel bestimmen
				float temp1 = fabs(coefficients->values[0] * 0.0 + coefficients->values[1] * 0.0 + coefficients->values[2] * 1.0);
				//float temp2 = sqrt((plane_coefficients[0][0] * plane_coefficients[0][0]) + (plane_coefficients[0][1] * plane_coefficients[0][1]) + (plane_coefficients[0][2] * plane_coefficients[0][2]));
				//float temp3 = sqrt((plane_coefficients[1][0] * plane_coefficients[1][0]) + (plane_coefficients[1][1] * plane_coefficients[1][1]) + (plane_coefficients[1][2] * plane_coefficients[1][2]));
				//float winkel = (acos(temp1 / (temp2 * temp3))) * 180 / M_PI;
				winkel = (acos(static_cast<double>(temp1)));
				std::cout << "Winkel zwischen z Achse und Bodennormale betraegt  " << (winkel * 180 / 3.14) << "grad" << std::endl;

				Eigen::Vector3d planeVector(coefficients->values[0],coefficients->values[1],coefficients->values[2]);
				Eigen::Vector3d zVector(0,0,1);
				Eigen::Vector3d normalVector = planeVector.cross(zVector);

				// Rotationsmatrix fuer Winkel um normalVector bestimmen
				Eigen::Affine3d m;
				m = Eigen::AngleAxisd(winkel, normalVector);
				Eigen::Matrix4f rotat = m.matrix().cast<float>();

				pcl::transformPointCloud (*cloud, *cloud, rotat);
				pcl::transformPointCloud (*clicked_points, *clicked_points, rotat);
			}
			counter++;
		}
		winkel = 180.0;

			float min= 9999.9;
			for(size_t i = 0; i < cloud->points.size(); i++){
				if (cloud->points[i].z < min){
					min = cloud->points[i].z;
				}
			}
			min += 0.2;
			for(size_t i = 0; i < cloud->points.size(); i++){
				cloud->points[i].z -= min;
			}

		pcl::copyPointCloud(*cloud, *unmodified_original_cloud);
		global_transformation_matrix = Eigen::Matrix4f::Identity();
		// alle ausgewählten Punkte aus Viewer entfernen
  		viewer->removePointCloud("clicked_points");
  		// alle ausgewählten Punkte löschen
		clicked_points->points.clear();
	}

	 //Event bei Numpad Enter - speichern in einem .pcd file
 	if (event.getKeySym () == "KP_Enter" && event.keyDown ())
 	{
 		// speichern, bei Fehler Ausgabe
 		if( pcl::io::savePCDFileASCII (saveDestination, *cloud) < 0)
 		{
 			printf("Saving cloud failed!");
 		}
 	}

	//Event bei minus - Punktgröße verkleinern
	if (event.getKeySym() == "minus" && event.keyDown()) {
		if(point_size >= 2){
			point_size--;
		}
	}

	//Event bei plus - Punktgröße verkleinern
	if (event.getKeySym() == "plus" && event.keyDown()) {
			point_size++;
	}

	//Event bei m - Modus zwischen ausschneiden und transformieren wechseln
	if (event.getKeySym() == "m" && event.keyDown()) {
		viewer->removeShape("a");
		viewer->removeShape("b");
		viewer->removeShape("c");
		viewer->removeShape("d");
		viewer->removeShape("e");
		viewer->removeShape("f");
		viewer->removeShape("g");
		viewer->removeShape("mode");
		viewer->removeShape("stepsize");

		if (mode == 0){
			mode = 1;
			viewer->addText("modus: cutter", 10, 10, 16, 0.8, 0.0, 0.0, "mode");
			std::stringstream stepsizeStream;
			stepsizeStream << "stepsize: " << schrittweite;
			viewer->addText(stepsizeStream.str(), 10, 30, 16, 0.8, 0.0, 0.0, "stepsize");
			if (!legendOFF){
				viewer->addText("numpad - enlarge cube", 10, 320, 16, 0.8, 0.1, 0.1, "a");
				viewer->addText("shift+numpad - shrink cube", 10, 300, 16, 0.8, 0.1, 0.1, "b");
				viewer->addText("numpad 5 - shrink stepsize", 10, 280, 16, 0.8, 0.1, 0.1, "c");
				viewer->addText("shift+numpad 5 - enlarge stepsize", 10, 260, 16, 0.8, 0.1, 0.1, "d");
				viewer->addText("b - remove all points outside of the cube", 10, 240, 16, 0.8, 0.1, 0.1, "e");
				viewer->addText("B - remove all points within the cube", 10, 220, 16, 0.8, 0.1, 0.1, "f");
				viewer->addText("shift+click - adjust borders", 10, 200, 16, 0.8, 0.1, 0.1, "g");
			}
		}else{
			if (mode == 1){
				mode = 0;
				viewer->addText("modus: transform", 10, 10, 16, 0.8, 0.0, 0.0, "mode");
				std::stringstream stepsizeStream;
				stepsizeStream << "stepsize: " << schrittweite;
				viewer->addText(stepsizeStream.str(), 10, 30, 16, 0.8, 0.0, 0.0, "stepsize");
				if (!legendOFF){
					viewer->addText("numpad - move points", 10, 320, 16, 0.8, 0.1, 0.1, "a");
					viewer->addText("shift+numpad - rotate points", 10, 300, 16, 0.8, 0.1, 0.1, "b");
					viewer->addText("numpad 5 - shrink stepsize", 10, 280, 16, 0.8, 0.1, 0.1, "c");
					viewer->addText("shift+numpad 5 - enlarge stepsize", 10, 260, 16, 0.8, 0.1, 0.1, "d");
					viewer->addText("t or T - print transform", 10, 240, 16, 0.8, 0.1, 0.1, "e");
					viewer->addText("X - type in transform", 10, 220, 16, 0.8, 0.1, 0.1, "f");
					viewer->addText("x - transform points using input transform", 10, 200, 16, 0.8, 0.1, 0.1, "g");
				}
			}
		}
	}

	//Event bei l - Legende in viewer an und ausschalten
	if (event.getKeySym() == "l" && event.keyDown()) {
		// legende anschalten
		if (legendOFF){
			// cutter
			viewer->removeShape("a");
			viewer->removeShape("b");
			viewer->removeShape("c");
			viewer->removeShape("d");
			viewer->removeShape("e");
			viewer->removeShape("f");
			viewer->removeShape("g");
			if (mode == 1){
				viewer->addText("numpad - enlarge cube", 10, 320, 16, 0.8, 0.1, 0.1, "a");
				viewer->addText("shift+numpad - shrink cube", 10, 300, 16, 0.8, 0.1, 0.1, "b");
				viewer->addText("numpad 5 - shrink stepsize", 10, 280, 16, 0.8, 0.1, 0.1, "c");
				viewer->addText("shift+numpad 5 - enlarge stepsize", 10, 260, 16, 0.8, 0.1, 0.1, "d");
				viewer->addText("b - remove all points outside of the cube", 10, 240, 16, 0.8, 0.1, 0.1, "e");
				viewer->addText("B - remove all points within the cube", 10, 220, 16, 0.8, 0.1, 0.1, "f");
				viewer->addText("shift+click - adjust borders", 10, 200, 16, 0.8, 0.1, 0.1, "g");
			}else{
				// transform
				viewer->addText("numpad - move points", 10, 320, 16, 0.8, 0.1, 0.1, "a");
				viewer->addText("shift+numpad - rotate points", 10, 300, 16, 0.8, 0.1, 0.1, "b");
				viewer->addText("numpad 5 - shrink stepsize", 10, 280, 16, 0.8, 0.1, 0.1, "c");
				viewer->addText("shift+numpad 5 - enlarge stepsize", 10, 260, 16, 0.8, 0.1, 0.1, "d");
				viewer->addText("t or T - print transform", 10, 240, 16, 0.8, 0.1, 0.1, "e");
				viewer->addText("X - type in transform", 10, 220, 16, 0.8, 0.1, 0.1, "f");
				viewer->addText("x - transform points using input transform", 10, 200, 16, 0.8, 0.1, 0.1, "g");
			}
			viewer->addText("delete - unpick all points", 15, 170, 16, 0.8, 0.1, 0.1, "delete");
			viewer->addText("n - orient cloud (pick one point on the floor first)", 15, 150, 16, 0.8, 0.1, 0.1, "orient");
			viewer->addText("backspace - \"undo\" ", 15, 130, 16, 0.8, 0.1, 0.1, "undo");
			viewer->addText("numpad enter - save cloud", 15, 110, 16, 0.8, 0.1, 0.1, "save");
			viewer->addText("m - switch modes", 15, 90, 16, 0.8, 0.1, 0.1, "m");
			viewer->addText("h - more details in terminal", 20, 70, 12, 0.8, 0.1, 0.1, "h");

			legendOFF = false;
		}else{
			viewer->removeShape("a");
			viewer->removeShape("b");
			viewer->removeShape("c");
			viewer->removeShape("d");
			viewer->removeShape("e");
			viewer->removeShape("f");
			viewer->removeShape("g");
			viewer->removeShape("h");
			viewer->removeShape("m");
			viewer->removeShape("save");
			viewer->removeShape("undo");
			viewer->removeShape("orient");
			viewer->removeShape("delete");

			legendOFF = true;
		}
	}

	if ( colorize ){
		colorHeight(cloud);
	}

	// cloud visualisieren und einfärben
    	pcl::visualization::PointCloudColorHandlerRGBField<pcl::PointXYZRGB> cloud_color_handler (cloud);
	viewer->updatePointCloud (cloud, cloud_color_handler, "original_cloud");
	viewer->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, point_size, "original_cloud");
}

// Regelt Keyboard und PointPicking Events 
boost::shared_ptr<pcl::visualization::PCLVisualizer> interactionCustomizationVis ()
{
	// deklariert viewer 
  	boost::shared_ptr<pcl::visualization::PCLVisualizer> viewer (new pcl::visualization::PCLVisualizer ("3D Viewer"));
  	
	// Setze Hintergrundfarbe und füge ein Koordinatensystem ein 
	viewer->setBackgroundColor (background_r, background_g, background_b);
  	viewer->addCoordinateSystem (1.0);
	
  	viewer->registerKeyboardCallback (keyboardEventOccurred, (void*)&viewer);

	viewer->registerPointPickingCallback (pp_callback, (void*)&viewer);

	if ( colorize ){
		colorHeight(unmodified_original_cloud);
	}
	pcl::copyPointCloud(*unmodified_original_cloud, *cloud);

	// cloud visualisieren und einfärben
    	pcl::visualization::PointCloudColorHandlerRGBField<pcl::PointXYZRGB> cloud_color_handler (cloud);
	viewer->addPointCloud (cloud, cloud_color_handler, "original_cloud");
	viewer->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, point_size, "original_cloud");

	pcl::PointXYZRGB point;
	point.x = -50;
	point.y = -50;
	point.z = -50;

	viewer->addText("l - legend", 20, 50, 12, 0.0, 0.0, 0.0, "legend");
	viewer->addText("modus: transform", 10, 10, 16, 0.8, 0.0, 0.0, "mode");
	std::stringstream stepsizeStream;
	stepsizeStream << "stepsize: " << schrittweite;
	viewer->addText(stepsizeStream.str(), 10, 30, 16, 0.8, 0.0, 0.0, "stepsize");

	pcl::copyPointCloud(*cloud, *backup_cloud);

  	return (viewer);
}



int main (int argc, char** argv)
{
	timer.reset();
	// Zeige Hilfe bei Eingabe von -h oder --help
	if (pcl::console::find_switch (argc, argv, "-h") || pcl::console::find_switch (argc, argv, "--help")) 
	{
		showHelp (argv[0]);
		return 0;
	}

	// Zeige Hilfe bei Eingabe von -h oder --help
	if (pcl::console::find_switch (argc, argv, "-nc"))
	{
		colorize = false;
	}

	// Namen der pcd Dateien aus Argumenten extrahieren | Funktioniert mit pcd (und ply Dateien)
	std::vector<int> filenames;	
	filenames = pcl::console::parse_file_extension_argument (argc, argv, ".pcd");
	
	// falls keine Dateie angegeben wurde, Hilfe anzeigen
	if (filenames.size () != 1) {
		showHelp (argv[0]);
		return -1;
	} 

	// Lade Dateien | Funktioniert mit pcd (und ply Dateien)
	// Lade cloud, falls Fehler zeige Hilfe
	if (pcl::io::loadPCDFile (argv[filenames[0]], *unmodified_original_cloud) < 0)
	{
		std::cout << "Fehler beim Laden der cloud. " << argv[filenames[0]] << std::endl << std::endl;
		showHelp (argv[0]);
		return -1;
	}

	// Name der Ausgabedatei festlegen und Speicherort als Inputordner festlegen
	saveDestination = argv[filenames[0]];
 	// Präfix "transformed_" vor Dateinamen der scene-Cloud ergänzen 
 	saveDestination.replace(saveDestination.find_last_of("/"),1,"/transformed_");

	// viewer Pointer initialisieren
	boost::shared_ptr<pcl::visualization::PCLVisualizer> viewer;
	// Abfrage von Events
	viewer = interactionCustomizationVis();

	// Endlosschleife solange Viewer nicht geschlossen wird
	while (!viewer->wasStopped ())
  	{
		// regelt Framerate und schützt ein wenig vor Überlastung, falls eine Taste zu lange gedrückt wird 
    	viewer->spinOnce (10);
    	boost::this_thread::sleep (boost::posix_time::microseconds (100000));
  	}

	return 0;
}
